/*
 * ProFTPD: mod_rusage -- a module for monitoring system resource usage
 *                        by a ProFTPD session process
 *
 * Copyright (c) 2011 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, TJ Saunders and other respective copyright holders
 * give permission to link this program with OpenSSL, and distribute the
 * resulting executable, without including the source code for OpenSSL in the
 * source distribution.
 *
 * This is mod_rusage, contrib software for proftpd 1.3.x and above.
 *
 * $Id: mod_rusage.c,v 1.2 2011/04/06 02:50:45 tj Exp tj $
 */

#include "mod_rusage.h"

/* If we are on a system which doesn't even have getrusage(2), we have
 * a problem.
 */
#ifndef HAVE_GETRUSAGE
# error "getrusage(2) support required"
#endif

/* If not explicitly configured, check the configured thresholds (if any)
 * every 15 seconds.
 */
#define RUSAGE_DEFAULT_INTERVAL		15

/* If not explicitly configured, log any exceeded thresholds at WARNING */
#define RUSAGE_DEFAULT_LOG_LEVEL	PR_LOG_WARNING

module rusage_module;

static int rusage_logfd = -1;
static int rusage_timerno = -1;

#define RUSAGE_ACTION_FL_LOG		3
#define RUSAGE_ACTION_FL_TERMINATE	4

static const char *trace_channel = "rusage";

/*  getrusage(2)
 *  sysctl(3)		FreeBSD, Mac OSX
 *  sysinfo(2)		Linux
 */

/* On Mac OSX, the ru_maxrss is in bytes, not KB.
 *
 *   http://lists.apple.com/archives/darwin-kernel/2009/Mar/msg00005.html
 *
 * On FreeBSD, ru_maxrss is in terms of memory pages:
 *
 *  http://www.khmere.com/freebsd_book/html/ch07.html
 *
 * Watch out for ru_maxrss negative values:
 *
 *  http://lists.freebsd.org/pipermail/freebsd-current/2006-June/063744.html
 *
 * On Linux, ru_maxrss is not always implemented:
 *
 *  http://lists.freebsd.org/pipermail/freebsd-hackers/2007-August/021605.html
 *
 * And on Solaris, it may not be implemented, either.
 *
 * Various tricks/attempts/discussion:
 *
 *   http://stackoverflow.com/questions/63166/how-to-determine-cpu-and-memory-consumption-from-inside-a-process
 *   http://www.perlmonks.org/?node_id=626693
 */

static void rusage_handle_time_thresholds(pr_table_t *tab, struct rusage *ru,
    unsigned int action_flag, int log_level) {
  unsigned long user_nsecs = 0, sys_nsecs = 0;
  void *v;
  const char *client_ip, *server_ip, *user, *cmd_phase = "(unknown)";
  int server_port;

  v = pr_table_get(tab, "UserTime", NULL);
  if (v != NULL) {
    user_nsecs = *((unsigned long *) v);
  }

  v = pr_table_get(tab, "SystemTime", NULL);
  if (v != NULL) {
    sys_nsecs = *((unsigned long *) v);
  }

  client_ip = pr_netaddr_get_ipstr(session.c->remote_addr);
  server_ip = pr_netaddr_get_ipstr(session.c->local_addr);
  server_port = pr_netaddr_get_port(session.c->local_addr);
  user = session.user ? session.user : "<unauthenticated>";

  switch (session.curr_phase) {
    case PRE_CMD:
      cmd_phase = "PRE_CMD";
      break;

    case CMD:
      cmd_phase = "CMD";
      break;

    case POST_CMD:
      cmd_phase = "POST_CMD";
      break;

    case POST_CMD_ERR:
      cmd_phase = "POST_CMD_ERR";
      break;

    case LOG_CMD:
      cmd_phase = "LOG_CMD";
      break;

    case LOG_CMD_ERR:
      cmd_phase = "LOG_CMD_ERR";
      break;
  }

  if (user_nsecs > 0 &&
      sys_nsecs > 0) {
    if (ru->ru_utime.tv_sec >= user_nsecs &&
        ru->ru_stime.tv_sec >= sys_nsecs) {
      switch (action_flag) {
        case RUSAGE_ACTION_FL_LOG:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold (used %lu secs, "
            "threshold %lu secs) and SystemTime threshold (used %lu secs, "
            "threshold %lu secs", client_ip, server_ip, server_port,
            user, cmd_phase, session.curr_cmd,
            (unsigned long) ru->ru_utime.tv_sec, user_nsecs,
            (unsigned long) ru->ru_stime.tv_sec, sys_nsecs);
          pr_log_pri(log_level, MOD_RUSAGE_VERSION
            ": client %s, server %s:%d, user '%s', command %s %s): "
            "process reached thresholds", client_ip, server_ip, server_port,
            user, cmd_phase, session.curr_cmd);
          break;

        case RUSAGE_ACTION_FL_TERMINATE:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold (used %lu secs, "
            "threshold %lu secs) and SystemTime threshold (used %lu secs, "
            "threshold %lu secs, terminating", client_ip, server_ip,
            server_port, user, cmd_phase, session.curr_cmd,
            (unsigned long) ru->ru_utime.tv_sec, user_nsecs,
            (unsigned long) ru->ru_stime.tv_sec, sys_nsecs);
          pr_log_pri(PR_LOG_WARNING, MOD_RUSAGE_VERSION
            ": (client %s, server %s:%d, user '%s', command %s %s): "
            "process reached thresholds, terminating", client_ip, server_ip,
            server_port, user, cmd_phase, session.curr_cmd);
#if PROFTPD_VERSION_NUMBER >= 0x0001030402
          pr_session_disconnect(&rusage_module,
            PR_SESS_DISCONNECT_BY_APPLICATION,
            "process time thresholds reached");
#else
          end_login(1);
#endif /* PROFTPD_VERSION_NUMBER less than 1.3.4rc2 */

          break;
      }
    }

  } else if (user_nsecs > 0) {
    if (ru->ru_utime.tv_sec >= user_nsecs) {
      switch (action_flag) {
        case RUSAGE_ACTION_FL_LOG:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold (used %lu secs, "
            "threshold %lu secs)", client_ip, server_ip, server_port, user,
            cmd_phase, session.curr_cmd, (unsigned long) ru->ru_utime.tv_sec,
            user_nsecs);
          pr_log_pri(log_level, MOD_RUSAGE_VERSION
            ": (client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold (used %lu secs, "
            "threshold %lu secs)", client_ip, server_ip, server_port, user,
            cmd_phase, session.curr_cmd, (unsigned long) ru->ru_utime.tv_sec,
            user_nsecs);
          break;

        case RUSAGE_ACTION_FL_TERMINATE:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold (used %lu secs, "
            "threshold %lu secs), terminating", client_ip, server_ip,
            server_port, user, cmd_phase, session.curr_cmd,
            (unsigned long) ru->ru_utime.tv_sec, user_nsecs);
          pr_log_pri(PR_LOG_WARNING, MOD_RUSAGE_VERSION
            ": (client %s, server %s:%d, user '%s', command %s %s): "
            "process reached UserTime threshold, terminating", client_ip,
            server_ip, server_port, user, cmd_phase, session.curr_cmd);
#if PROFTPD_VERSION_NUMBER >= 0x0001030402
          pr_session_disconnect(&rusage_module,
            PR_SESS_DISCONNECT_BY_APPLICATION,
            "process time threshold reached");
#else
          end_login(1);
#endif /* PROFTPD_VERSION_NUMBER less than 1.3.4rc2 */

          break;
      }
    }

  } else if (sys_nsecs > 0) {
    if (ru->ru_stime.tv_sec >= sys_nsecs) {
      switch (action_flag) {
        case RUSAGE_ACTION_FL_LOG:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached SystemTime threshold (used %lu secs, "
            "threshold %lu secs)", client_ip, server_ip, server_port, user,
            cmd_phase, session.curr_cmd, (unsigned long) ru->ru_stime.tv_sec,
            sys_nsecs);
          pr_log_pri(log_level, MOD_RUSAGE_VERSION
            ": (client %s, server %s:%d, user '%s', command %s %s): "
            "process reached SystemTime threshold (used %lu secs, "
            "threshold %lu secs)", client_ip, server_ip, server_port, user,
            cmd_phase, session.curr_cmd, (unsigned long) ru->ru_stime.tv_sec,
            sys_nsecs);
          break;

        case RUSAGE_ACTION_FL_TERMINATE:
          (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
            "(client %s, server %s:%d, user '%s', command %s %s): "
            "process reached SystemTime threshold (used %lu secs, "
            "threshold %lu secs), terminating", client_ip, server_ip,
            server_port, user, cmd_phase, session.curr_cmd,
            (unsigned long) ru->ru_stime.tv_sec, sys_nsecs);
          pr_log_pri(PR_LOG_WARNING, MOD_RUSAGE_VERSION
            ": (client %s, server %s:%d, user '%s', command %s %s): "
            "process reached SystemTime threshold, terminating",
            client_ip, server_ip, server_port, user, cmd_phase,
            session.curr_cmd);
#if PROFTPD_VERSION_NUMBER >= 0x0001030402
          pr_session_disconnect(&rusage_module,
            PR_SESS_DISCONNECT_BY_APPLICATION,
            "process time threshold reached");
#else
          end_login(1);
#endif /* PROFTPD_VERSION_NUMBER less than 1.3.4rc2 */

          break;
      }
    }
  }
}

/* Configuration handlers
 */

/* usage: RUsageEngine on|off */
MODRET set_rusageengine(cmd_rec *cmd) {
  config_rec *c;
  int bool = -1;

  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL);

  bool = get_boolean(cmd, 1);
  if (bool == -1) {
    CONF_ERROR(cmd, "expected Boolean parameter");
  }
 
  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(int));
  *((int *) c->argv[0]) = bool;

  return PR_HANDLED(cmd);
}

/* usage: RUsageInterval seconds */
MODRET set_rusageinterval(cmd_rec *cmd) {
  int nsecs = 0;
  config_rec *c;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  nsecs = atoi(cmd->argv[1]);
  if (nsecs <= 0) {
    CONF_ERROR(cmd, "must be a positive number");
  }

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(int));
  *((int *) c->argv[0]) = nsecs;

  return PR_HANDLED(cmd);
}

/* usage: RUsageLog path|"none" */
MODRET set_rusagelog(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  (void) add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
  return PR_HANDLED(cmd);
}

/* usage: RUsageThreshold action threshold-type threshold-value ... */
MODRET set_rusagethreshold(cmd_rec *cmd) {
  register unsigned int i;
  config_rec *c;
  unsigned int action_flag;
  int log_level = RUSAGE_DEFAULT_LOG_LEVEL;
  pr_table_t *tab;

  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  if (cmd->argc < 4) {
    CONF_ERROR(cmd, "Wrong number of parameters");

  } else {
    int npairs;

    /* Make sure we have an even number of args for the key/value pairs. */ 

    npairs = cmd->argc - 2;
    if (npairs % 2 != 0) {
      CONF_ERROR(cmd, "Wrong number of parameters");
    }
  }

  /* "action" can be either:
   *
   *  log[:level]
   *  terminate
   */

  if (strncmp(cmd->argv[1], "log:", 4) == 0) {
    int level;

    action_flag = RUSAGE_ACTION_FL_LOG;

    level = pr_log_str2sysloglevel(cmd->argv[1] + 4);
    if (level < 0) {
      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "bad RUsageThreshold log level: ",
        cmd->argv[1] + 4, NULL));
    }

    log_level = level;

  } else if (strncmp(cmd->argv[1], "log", 4) == 0) {
    action_flag = RUSAGE_ACTION_FL_LOG;

  } else if (strncmp(cmd->argv[1], "terminate", 10) == 0) {
    action_flag = RUSAGE_ACTION_FL_TERMINATE;

  } else {
    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "bad RUsageThreshold action :",
      cmd->argv[1], NULL));
  }

  c = add_config_param(cmd->argv[0], 3, NULL, NULL, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(unsigned int));
  *((unsigned int *) c->argv[0]) = action_flag;

  if (action_flag == RUSAGE_ACTION_FL_LOG) {
    c->argv[1] = pcalloc(c->pool, sizeof(int));
    *((int *) c->argv[1]) = log_level;
  }

  tab = pr_table_alloc(c->pool, 0);
  c->argv[2] = tab;

  for (i = 2; i < cmd->argc; i++) {
    if (strncmp(cmd->argv[i], "UserTime", 9) == 0) {
      unsigned long nsecs;
      char *tmp = NULL;
      void *v;

      nsecs = strtoul(cmd->argv[i+1], &tmp, 10);
      if (tmp && *tmp) {
        CONF_ERROR(cmd, "badly formatted UserTime threshold value");
      }

      v = palloc(c->pool, sizeof(unsigned long));
      *((unsigned long *) v) = nsecs;

      if (pr_table_add(tab, pstrdup(c->pool, cmd->argv[i]), v,
          sizeof(unsigned long)) < 0) {
        CONF_ERROR(cmd, pstrcat(cmd->tmp_pool,
          "error storing 'UserTime' threshold value: ", strerror(errno),
          NULL));
      }

      /* Don't forget to advance i past the value. */
      i++;

    } else if (strncmp(cmd->argv[i], "SystemTime", 11) == 0) {
      unsigned long nsecs;
      char *tmp = NULL;
      void *v;

      nsecs = strtoul(cmd->argv[i+1], &tmp, 10);
      if (tmp && *tmp) {
        CONF_ERROR(cmd, "badly formatted SystemTime threshold value");
      }

      v = palloc(c->pool, sizeof(unsigned long));
      *((unsigned long *) v) = nsecs;

      if (pr_table_add(tab, pstrdup(c->pool, cmd->argv[i]), v,
          sizeof(unsigned long)) < 0) {
        CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, 
          "error storing 'SystemTime' threshold value: ", strerror(errno),
          NULL));
      }

      /* Don't forget to advance i past the value. */
      i++;

    } else {
      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool,
        ": unknown RUsageThreshold type: '", cmd->argv[i], "'", NULL));
    }
  }

  return PR_HANDLED(cmd);
}

/* Timer handlers
 */

static int rusage_timer_cb(CALLBACK_FRAME) {
  int res;
  struct rusage ru;
  config_rec *c;

  res = getrusage(RUSAGE_SELF, &ru);
  if (res < 0) {
    (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
      "error calling getrusage(2): %s", strerror(errno));
    return 0;
  }

  if (pr_trace_get_level(trace_channel) >= 18) {
    pr_trace_msg(trace_channel, 18,
      "platform %s: ru_utime: %lu tv_sec, %lu tv_usec", PR_PLATFORM,
      (unsigned long) ru.ru_utime.tv_sec, (unsigned long) ru.ru_utime.tv_usec);

    pr_trace_msg(trace_channel, 18,
      "platform %s: ru_stime: %lu tv_sec, %lu tv_usec", PR_PLATFORM,
      (unsigned long) ru.ru_stime.tv_sec, (unsigned long) ru.ru_stime.tv_usec);

    /* Other than ru_utime and ru_stime, the struct rusage fields are NOT
     * mandated by POSIX.  Thus some platforms don't populate these fields.
     */

    pr_trace_msg(trace_channel, 18, "platform %s: ru_maxrss: %ld",
      PR_PLATFORM, ru.ru_maxrss);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_ixrss: %ld",
      PR_PLATFORM, ru.ru_ixrss);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_idrss: %ld",
      PR_PLATFORM, ru.ru_idrss);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_isrss: %ld",
      PR_PLATFORM, ru.ru_isrss);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_minflt: %ld",
      PR_PLATFORM, ru.ru_minflt);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_majflt: %ld",
      PR_PLATFORM, ru.ru_majflt);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_nswap: %ld",
      PR_PLATFORM, ru.ru_nswap);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_inblock: %ld",
      PR_PLATFORM, ru.ru_inblock);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_oublock: %ld",
      PR_PLATFORM, ru.ru_oublock);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_msgsnd: %ld",
      PR_PLATFORM, ru.ru_msgsnd);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_msgrcv: %ld",
      PR_PLATFORM, ru.ru_msgrcv);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_nsignals: %ld",
      PR_PLATFORM, ru.ru_nsignals);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_nvcsw: %ld",
      PR_PLATFORM, ru.ru_nvcsw);
    pr_trace_msg(trace_channel, 18, "platform %s: ru_nivcsw: %ld",
      PR_PLATFORM, ru.ru_nivcsw);
  }
   
  c = find_config(main_server->conf, CONF_PARAM, "RUsageThreshold", FALSE);
  while (c) {
    unsigned int action_flag;
    int log_level = RUSAGE_DEFAULT_LOG_LEVEL;
    pr_table_t *tab;

    pr_signals_handle();

    action_flag = *((unsigned int *) c->argv[0]);
    if (action_flag == RUSAGE_ACTION_FL_LOG) {
      log_level = *((int *) c->argv[1]);
    }

    tab = c->argv[2];

    /* Check configured thresholds against rusage here */
    rusage_handle_time_thresholds(tab, &ru, action_flag, log_level);

    c = find_config_next(c, c->next, CONF_PARAM, "RUsageThreshold", FALSE);
  }

  /* Make sure we are called again. */
  return 1;
}

/* Event handlers
 */

#ifdef PR_SHARED_MODULE
static void rusage_mod_unload_ev(const void *event_data, void *user_data) {
  if (strncmp((const char *) event_data, "mod_rusage.c", 13) == 0) {
    pr_event_unregister(&rusage_module, NULL, NULL);
  }
}
#endif /* PR_SHARED_MODULE */

/* Initialization routines
 */

static int rusage_init(void) {

#ifdef PR_SHARED_MODULE
  pr_event_register(&rusage_module, "core.module-unload", rusage_mod_unload_ev,
    NULL);
#endif
  return 0;
}

static int rusage_sess_init(void) {
  config_rec *c;
  int engine = FALSE;
  int interval = RUSAGE_DEFAULT_INTERVAL;

  c = find_config(main_server->conf, CONF_PARAM, "RUsageEngine", FALSE);
  if (c) {
    engine = *((int *) c->argv[0]);
  }

  if (!engine) {
    return 0;
  }

  c = find_config(main_server->conf, CONF_PARAM, "RUsageLog", FALSE);
  if (c) {
    char *path;

    path = c->argv[0];
    if (strncmp(path, "none", 5) != 0) {
      int res;
    
      pr_signals_block();
      PRIVS_ROOT
      res = pr_log_openfile(path, &rusage_logfd, 0660);
      PRIVS_RELINQUISH
      pr_signals_unblock();

      if (res < 0) {
        if (res == -1) {
          pr_log_pri(PR_LOG_NOTICE, MOD_RUSAGE_VERSION
            ": notice: unable to open RUsageLog '%s': %s", path,
            strerror(errno));

        } else if (res == PR_LOG_WRITABLE_DIR) {
          pr_log_pri(PR_LOG_NOTICE, MOD_RUSAGE_VERSION
            ": notice: unable to open RUsageLog '%s': parent directory is "
            "world-writable", path);

        } else if (res == PR_LOG_SYMLINK) {
          pr_log_pri(PR_LOG_NOTICE, MOD_RUSAGE_VERSION
            ": notice: unable to open RUsageLog '%s': cannot log to a symlink",
            path);
        }
      }
    }
  }

  /* Make sure we have at least one threshold configured. */
  c = find_config(main_server->conf, CONF_PARAM, "RUsageThreshold", FALSE);
  if (c == NULL) {
    (void) pr_log_writefile(rusage_logfd, MOD_RUSAGE_VERSION,
      "%s", "no RUsageThreshold configured, disabling module");
    return 0;
  }

  c = find_config(main_server->conf, CONF_PARAM, "RUsageInterval", FALSE);
  if (c) {
    interval = *((int *) c->argv[0]);
  }

  rusage_timerno = pr_timer_add(interval, -1, &rusage_module, rusage_timer_cb,
    "getrusage(2) poll");
  return 0;
}

/* Module API tables
 */

static conftable rusage_conftab[] = {
  { "RUsageEngine",	set_rusageengine,	NULL },
  { "RUsageInterval",	set_rusageinterval,	NULL },
  { "RUsageLog",	set_rusagelog,		NULL },
  { "RUsageThreshold",	set_rusagethreshold,	NULL },
  { NULL }
};

module rusage_module = {
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* Module name */
  "rusage",

  /* Module configuration handler table */
  rusage_conftab,

  /* Module command handler table */
  NULL,

  /* Module authentication handler table */
  NULL,

  /* Module initialization function */
  rusage_init,

  /* Session initialization function */
  rusage_sess_init,

  /* Module version */
  MOD_RUSAGE_VERSION
};
